// This may look like C code, but it is really -*- C++ -*-
/*
    newtonsc NewtonScript compiler
    Copyright (C) 1995  Ross Sponholtz

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.



*/

/*
   This file implements newtonscript semantic objects
*/


#ifndef PROD_H_INC
#define PROD_H_INC


#include "objects.h"
#include "evalue.h"
#include "codebloc.h"

extern int lineno;


class Production
{
 public:
  int lineNo;
  void outtab(int c)
    {
      int i;
      
      cout << endl;
      cout.form("%5d",lineNo);
      for (i=0;i<c;++i)
	cout << " ";
    }

};


class Constituent  : public Production
{
private:
  int nothing; 
public:
  virtual int output(int tc) = 0;
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer) = 0;
  virtual int walk(Symtab *context) = 0;
};


typedef DLList<Constituent *> ConstituentList;

class Input : public Production
{
  ConstituentList theConstituents;
public:
  Input(Constituent *c)
    {
      theConstituents.append(c);
      lineNo = lineno;
    }
  Input *AddConstituent(Constituent *c)
    {
      theConstituents.append(c);
      return this;

    }
  Input *PrependConstituent(Constituent *c)
    {
      theConstituents.prepend(c);
      return this;
    }

  virtual int output(int tc);
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer);
  virtual int walk(Symtab *context);
};


class Expression : public Constituent
{
public:
  virtual int output(int tc) = 0;
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer) = 0;
  virtual int walk(Symtab *context) = 0;
  virtual EValue evalExpr() = 0;
};

typedef DLList<Expression *> ExpressionList;

class ExpressionSequence : public Production
{
  ExpressionList theExpressions;
public:
  ExpressionSequence(Expression *c)
    {
      theExpressions.append(c);
      lineNo = lineno;
    }

  ExpressionSequence *AddExpression(Expression *c)
    {
      theExpressions.append(c);
      return this;
    }
  ExpressionSequence *PrependExpression(Expression *c)
    {
      theExpressions.prepend(c);
      return this;
    }
  virtual int output(int tc);
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer);
  virtual int walk(Symtab *context);
  virtual EValue evalExpr()
    {EValue t; return t;}
  
};



class Lvalue : public Expression
{
public:
  virtual int output(int tc)= 0; 
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer) = 0;
  virtual int emitAssignTo(Codeblock *theCode) = 0;
  virtual int walk(Symtab *context) = 0;
  virtual EValue evalExpr() = 0;
};



class Symbol : public Lvalue
{
  String value;
  SymtabEntry *symbolTableRef;
public:
  Symbol(char *nv):value(nv),lineNo(lineno){}
  Symbol(String nv):value(nv),lineNo(lineno){}
  SetLocal(void)    {symbolTableRef->SetLocal();}
  SetGlobal(void)    {symbolTableRef->SetGlobal();}
  SetConstant(void)    {symbolTableRef->SetConstant();}
  SetOther(void)    {symbolTableRef->SetOther();}
  SetArg()    {symbolTableRef->SetArg();}
  Order() {return symbolTableRef->Order();}
  String SymbolValue() {return value;}
  SymtabEntry *GetRef() {return symbolTableRef;}
  virtual int output(int tc);
  virtual int dump();
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer);
  virtual int emitAssignTo(Codeblock *theCode);
  int emitIncr(Codeblock *theCode);
  virtual int walk(Symtab *context);
  int MarkMagic();
  NO_Symbol *GetNOSymbol()
    {return symbolTableRef->getNOSymbol();}

  virtual EValue evalExpr()
    {EValue ttt;
     ttt=symbolTableRef->theValue;
     return ttt;};
};

typedef DLList<Symbol *> SymbolList;

class FormalArgumentList : public Production
{
  SymbolList theSymbols;
public:
  FormalArgumentList(Symbol *c)
    {
      theSymbols.append(c);
      lineNo = lineno;
    }

  FormalArgumentList *AddSymbol(Symbol *c)
    {
      theSymbols.append(c);
      return this;
    }
  FormalArgumentList *PrependSymbol(Symbol *c)
    {
      theSymbols.prepend(c);
      return this;
    }

  int countArgs()
    {
      return theSymbols.length();
    }
  AddArgsToFrame(NO_Frame *theFrame);

  virtual int output(int tc);
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer);
  virtual int walk(Symtab *context);
};

class GlobalDeclaration : public Constituent
{
public:
  virtual int output(int tc) = 0;
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer) = 0;
  virtual int walk(Symtab *context) = 0;
};



class InitClause : public GlobalDeclaration
{
  Symbol *theSymbol;
  Expression *theExpression;
public:
  InitClause(Symbol *s,Expression *e):theSymbol(s),theExpression(e),lineNo(lineno){}
  void MarkLocal(void) {theSymbol->SetLocal();}
  void MarkGlobal(void) {theSymbol->SetGlobal();}
  virtual int output(int tc);
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer);
  virtual int walk(Symtab *context);
};

typedef DLList<InitClause *> InitClauseList;


class GlobalFunctionDecl : public GlobalDeclaration
{
  int isGlobal;
  Symbol *theSymbol;
  FormalArgumentList *theArgs;
  Expression *theExpression;
public:
  GlobalFunctionDecl(int i,Symbol *s,FormalArgumentList *a,Expression *e):
    isGlobal(i),theSymbol(s),theArgs(a),theExpression(e),lineNo(lineno){}

  virtual int output(int tc);
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer);
  virtual int walk(Symtab *context);
};

class GlobalWithInitialization : public GlobalDeclaration
{
  InitClause *theInitClause;
public:
  GlobalWithInitialization(InitClause *i):theInitClause(i),lineNo(lineno){}
  virtual int output(int tc);
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer);
  virtual int walk(Symtab *context);
};



class SimpleExpression : public Expression
{
  Expression *left_expr;
  Expression *right_expr;
  int theOperator;
public:
  SimpleExpression(Expression *l,Expression *r,int o):left_expr(l),right_expr(r),theOperator(o),lineNo(lineno){}
  virtual int output(int tc);
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer);
  virtual int walk(Symtab *context);
  virtual EValue evalExpr();
};

class CompoundExpression : public Expression
{
  ExpressionSequence *theExpressions;
public:
  CompoundExpression(ExpressionSequence *es):theExpressions(es),lineNo(lineno){}
  virtual int output(int tc);
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer);
  virtual int walk(Symtab *context);
  virtual EValue evalExpr()
    {EValue t; return t;};
};


class Literal : public Expression
{
public:
  virtual int output(int tc) = 0;
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer) = 0;
  virtual int walk(Symtab *context) = 0;
  virtual EValue evalExpr() = 0;
};

class SimpleLiteral : public Literal
{
public:
  virtual int output(int tc) = 0;
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer) = 0;
  virtual int walk(Symtab *context) = 0;
  virtual EValue evalExpr() = 0;
};

class Object : public Literal
{
public:
  virtual int output(int tc) = 0;
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer) = 0;
  virtual int walk(Symtab *context) = 0;
  virtual EValue evalExpr() = 0;
};

typedef DLList<Object *> ObjectList;


class SimpleObject : public Object
{
  SimpleLiteral *theSimpleLiteral;
public:
  SimpleObject(SimpleLiteral *s):theSimpleLiteral(s),lineNo(lineno){}
  virtual int output(int tc);
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer);
  virtual int walk(Symtab *context);
  virtual EValue evalExpr()
     {return theSimpleLiteral->evalExpr();}
};


class PathExpression : public Production
{
  SymbolList theSymbols;
public:
  PathExpression(Symbol *s)
    {
      theSymbols.append(s);
      lineNo = lineno;
    }

  PathExpression *AddSymbol(Symbol *s)
    {
      theSymbols.append(s);
      return this;
    }

  //RWS fill in constructor
  virtual int output(int tc);
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer);
  virtual int walk(Symtab *context);
};


class PathObject : public Object
{
  PathExpression *theObject;
public:
  PathObject(PathExpression *p):theObject(p),lineNo(lineno){}
  virtual int output(int tc);
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer);
  virtual int walk(Symtab *context);
  virtual EValue evalExpr()
    {EValue t; return t;};
};


class Array : public Production
{
  Symbol *initialSymbol;
  ObjectList theObjects;
public:
  Array(Object *c)
    {
      theObjects.append(c);
      lineNo = lineno;
    }
  Array(Symbol *s)
    {
      initialSymbol = s;
      lineNo = lineno;
    }
  Array *SetSymbol(Symbol *s)
    {initialSymbol = s;
     return this;
    }

  Array *AddObject(Object *c)
    {
      theObjects.append(c);
      return this;
    }

  virtual int output(int tc);
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer);
  virtual int walk(Symtab *context);
};


class ArrayObject : public Object
{
  Array *theObject;
public:
  ArrayObject(Array *a):theObject(a),lineNo(lineno){}
  virtual int output(int tc);
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer);
  virtual int walk(Symtab *context);
  virtual EValue evalExpr()
    {EValue t; return t;};
};



class FrameSlot : public Production
{
  Symbol *theSymbol;
  Object *theObject;
public:
  FrameSlot(Symbol *s,Object *o):theSymbol(s),theObject(o),lineNo(lineno){}
  virtual int output(int tc);
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer);
  virtual int walk(Symtab *context);
};

typedef DLList<FrameSlot *> FrameSlotList;

class Frame : public Production
{
  FrameSlotList theSlots;
public:

  Frame(FrameSlot *c)
    {
      if (c != NULL)
	theSlots.append(c);
      lineNo = lineno;
    }
  Frame(void):lineNo(lineno){}

  Frame *AddSlot(FrameSlot *c)
    {
      theSlots.append(c);
      return this;
    }
  Frame *PrependSlot(FrameSlot *c)
    {
      theSlots.prepend(c);
      return this;
    }

  virtual int output(int tc);
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer);
  virtual int walk(Symtab *context);
};

class FrameObject : public Object
{
  Frame *theObject;
public:
  FrameObject(Frame *f):theObject(f),lineNo(lineno){}
  virtual int output(int tc);
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer);
  virtual int walk(Symtab *context);
  virtual EValue evalExpr()
    {EValue t; return t;};
};



class Constructor : public Expression
{
public:
  virtual int output(int tc) = 0;
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer) = 0;
  virtual int walk(Symtab *context) = 0;
  virtual EValue evalExpr()
    {EValue t; return t;};
};

class ArrayConstructor : public Constructor
{
  Symbol *initialSymbol;
  ExpressionList theExpressions;
public:
  ArrayConstructor(Expression *c)
    {
      initialSymbol = NULL;
      theExpressions.append(c);
      lineNo = lineno;
    }
  ArrayConstructor(Symbol *c)
    {
      initialSymbol = c;
      lineNo = lineno;
    }
  
  ArrayConstructor()
    {
      initialSymbol = NULL;
      lineNo = lineno;
    }
  
  ArrayConstructor *SetSymbol(Symbol *c)
    {
      initialSymbol = c;
      return this;
    }
  ArrayConstructor *AddExpression(Expression *c)
    {
      theExpressions.append(c);
      return this;
    }
  ArrayConstructor *PrependExpression(Expression *c)
    {
      theExpressions.prepend(c);
      return this;
    }
  virtual int output(int tc);
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer);
  virtual int walk(Symtab *context);
};


class FrameConstructorSlot : public Production
{
  Symbol *theSymbol;
  Expression *theExpression;
public:
  FrameConstructorSlot(Symbol *s,Expression *e):theSymbol(s),theExpression(e){lineNo = lineno;}
  virtual int output(int tc);
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer);
  virtual int walk(Symtab *context);
};

typedef DLList<FrameConstructorSlot *> FSSlotList;

class FrameConstructor : public Constructor
{
  FSSlotList theSlots;
public:

  FrameConstructor(void):lineNo(lineno)
    {}

  FrameConstructor(FrameConstructorSlot *c)
    {
      theSlots.append(c);
      lineNo = lineno;
    }

  FrameConstructor *AddSlot(FrameConstructorSlot *c)
    {
      theSlots.append(c);
      return this;
    }
  FrameConstructor *PrependSlot(FrameConstructorSlot *c)
    {
      theSlots.prepend(c);
      return this;
    }

  virtual int output(int tc);
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer);
  virtual int walk(Symtab *context);
};


class FunctionConstructor : public Constructor
{
  FormalArgumentList *theArgs;
  Expression *theExpression;
  Symtab theSymbolTable;


public:
  FunctionConstructor(FormalArgumentList *a,Expression *e,Symtab *pp)
    {
      theArgs=a;
      theExpression=e;
      lineNo=lineno;
      theSymbolTable.SetParent(pp);
      
    }
  virtual int output(int tc);
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer);
  virtual int walk(Symtab *context);
};


class FrameAccessor : public Lvalue
{
  Expression *leadingExpr;
  Symbol *theSymbol;
  Expression *trailingExpr;
public:
  FrameAccessor(Expression *l,Symbol *s,Expression *t):leadingExpr(l),theSymbol(s),trailingExpr(t),lineNo(lineno){}
  virtual int output(int tc);
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer);
  virtual int emitAssignTo(Codeblock *theCode);
  virtual int walk(Symtab *context);
  virtual EValue evalExpr()
    {EValue t; return t;};
};

class ArrayAccessor : public Lvalue
{
  Expression *leadingExpr;
  Expression *trailingExpr;
public:
  ArrayAccessor(Expression *l,Expression *t):leadingExpr(l),trailingExpr(t),lineNo(lineno){}
  virtual int output(int tc);
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer);
  virtual int emitAssignTo(Codeblock *theCode);
  virtual int walk(Symtab *context);
  virtual EValue evalExpr()
    {EValue t; return t;};
};


class Assignment : public Expression
{
  Lvalue *theLvalue;
  Expression *theExpression;
public:
  Assignment(Lvalue *l,Expression *e):theLvalue(l),theExpression(e),lineNo(lineno){}
  virtual int output(int tc);
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer);
  virtual int walk(Symtab *context);
  virtual EValue evalExpr()
    {EValue t; return t;};
};

class ExistsExpr : public Expression
{
public:
  virtual int output(int tc) = 0;
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer) = 0;
  virtual int walk(Symtab *context) = 0;
  virtual EValue evalExpr()
    {EValue t; return t;};
};

class SymbolExists : public ExistsExpr
{
  Symbol *theSymbol;
  int colspec;
public:
  SymbolExists(Symbol *s,int cs):theSymbol(s),colspec(cs),lineNo(lineno){}
  virtual int output(int tc);
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer);
  virtual int walk(Symtab *context);
};

class FrameAccExists : public ExistsExpr
{
  FrameAccessor *theFrameAcc;
public:
  FrameAccExists(FrameAccessor *f):theFrameAcc(f),lineNo(lineno){}
  virtual int output(int tc);
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer);
  virtual int walk(Symtab *context);
};

class ExprExists : public ExistsExpr
{
  Expression *theExpression;
  Symbol *theSymbol;
public:
  ExprExists(Expression *e,Symbol *s):theExpression(e),theSymbol(s),lineNo(lineno){}
  virtual int output(int tc);
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer);
  virtual int walk(Symtab *context);
};

class ActualArgumentList : public Production
{
  ExpressionList theArgs;
public:

  ActualArgumentList(Expression *c)
    {
      theArgs.append(c);
      lineNo = lineno;
    }

  ActualArgumentList *AddArg(Expression *c)
    {
      theArgs.append(c);
      return this;
    }
  ActualArgumentList *PrependArg(Expression *c)
    {
      theArgs.prepend(c);
      return this;
    }

  int countArgs()
    {
      return theArgs.length();
    }


  virtual int output(int tc);
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer);
  virtual int walk(Symtab *context);
};


class FunctionCall : public Expression
{
  int isCall;
  Symbol *theSymbol;
  Expression *theExpression;
  ActualArgumentList *theArgs;
public:
  FunctionCall(int ic,Symbol *s,Expression *e,ActualArgumentList *aa):isCall(ic),theSymbol(s),theExpression(e),theArgs(aa),lineNo(lineno){}
  virtual int output(int tc);
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer);
  virtual int walk(Symtab *context);
  virtual EValue evalExpr()
    {EValue t; return t;}
};


class MessageSend : public Expression
{
  int isInherited;
  Expression *theExpression;
  int col_or_cq;
  Symbol *theSymbol;
  ActualArgumentList *theArgs;
public:
  MessageSend(int ii,Expression *ee,int cq,Symbol *s,ActualArgumentList *aa):
  isInherited(ii),theExpression(ee),col_or_cq(cq),theSymbol(s),theArgs(aa),lineNo(lineno){}
  virtual int output(int tc);
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer);
  virtual int walk(Symtab *context);
  virtual EValue evalExpr()
    {EValue t; return t;}
};

class IfExpression : public Expression
{
  Expression *boolExpr;
  Expression *thenExpr;
  Expression *elseExpr;
public:
  IfExpression(Expression *b,Expression *t,Expression *e):boolExpr(b),thenExpr(t),elseExpr(e),lineNo(lineno){}
  virtual int output(int tc);
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer);
  virtual int walk(Symtab *context);
  virtual EValue evalExpr()
    {EValue t; return t;}
};


class Iteration : public Expression
{
  InstructionList theBreakInstrs;
public:
  addBreak(Instruction *ii) {theBreakInstrs.append(ii);}
  PatchBreaksTo(Instruction *p)
    {
      InstructionList tempList = theBreakInstrs;
      Instruction *tempItem;
  
      while (!tempList.empty())
	{
	  tempItem = tempList.remove_front();
	  tempItem->SetPatchTo(p);
	};
    }
  PatchBreaksAfter(Instruction *p)
    {
      InstructionList tempList = theBreakInstrs;
      Instruction *tempItem;
  
      while (!tempList.empty())
	{
	  tempItem = tempList.remove_front();
	  tempItem->SetPatchAfter(p);
	};
    }

  virtual int output(int tc) = 0;
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer) = 0;
  virtual int walk(Symtab *context) = 0;
  virtual EValue evalExpr()
    {EValue t; return t;};
};

class InfiniteLoop : public Iteration
{
  Expression *theExpression;
public:
  InfiniteLoop(Expression *ex):theExpression(ex),lineNo(lineno){}
  virtual int output(int tc);
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer);
  virtual int walk(Symtab *context);
};

class ForLoop : public Iteration
{
  Symbol *loopSymbol;
  Expression *initialExpression;
  Expression *toExpression;
  Expression *byExpression;
  Expression *doExpression;
  Symbol *limitSymbol;
  Symbol *incrSymbol;
public:
  ForLoop(
	  Symbol *ls,
	  Expression *in,
	  Expression *to,
	  Expression *by,
	  Expression *doe):
	    loopSymbol(ls),
	    initialExpression(in),
	    toExpression(to),
	    byExpression(by),
	    doExpression(doe),
	    lineNo(lineno) {}
  virtual int output(int tc);
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer);
  virtual int walk(Symtab *context);
};

class ForEachLoop : public Iteration
{
  Symbol *initialSymbol;
  Symbol *nextSymbol;
  int isDeeply;
  Expression *inExpression;
  int isDo;
  Expression *doExpression;

public:
  ForEachLoop(  Symbol *i,  Symbol *n,  int isd,  Expression *in,
  int isdo,  Expression *doexp):
    initialSymbol(i),
    nextSymbol(n),
    isDeeply(isd),
    inExpression(in),
    isDo(isdo),
    doExpression(doexp),
    lineNo(lineno) {}

  virtual int output(int tc);
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer);
  virtual int walk(Symtab *context);
};

class WhileLoop : public Iteration
{
  Expression *whileExpression;
  Expression *doExpression;
public:
  WhileLoop(Expression *we,Expression *de):whileExpression(we),doExpression(de),lineNo(lineno) {}
  virtual int output(int tc);
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer);
  virtual int walk(Symtab *context);
};

class RepeatLoop : public Iteration
{
  ExpressionSequence *theExpressions;
  Expression *untilExpression;
public:
  RepeatLoop(ExpressionSequence *es,Expression *ue):theExpressions(es),untilExpression(ue),lineNo(lineno) {}

  virtual int output(int tc);
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer);
  virtual int walk(Symtab *context);
};

class BreakExpr : public Expression
{
  Expression *theExpression;
public:
  BreakExpr(Expression *ex):theExpression(ex),lineNo(lineno) {}
  virtual int output(int tc);
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer);
  virtual int walk(Symtab *context);
  virtual EValue evalExpr()
    {EValue t; return t;};
};


class OnExceptionClause : public Production
{
  Symbol *theSymbol;
  Expression *theExpression;
public:
  OnExceptionClause(Symbol *ss,Expression *ee):theSymbol(ss),theExpression(ee),lineNo(lineno){}
  virtual int output(int tc);
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer);
  virtual int walk(Symtab *context);
};

typedef DLList<OnExceptionClause *> ExcClauseList;


class TryExpr : public Expression
{
  ExpressionSequence *theExpressions;
  ExcClauseList theClauses;
public:

  TryExpr(OnExceptionClause *c)
    {
      theClauses.append(c);
      theExpressions = NULL;
      lineNo = lineno;
    }

  TryExpr(ExpressionSequence *c)
    {
      theExpressions = c;
      lineNo = lineno;
    }

  TryExpr(ExpressionSequence *e,OnExceptionClause *c)
    {
      theClauses.append(c);
      theExpressions = e;
      lineNo = lineno;
    }

  TryExpr *SetExpr(ExpressionSequence *c)
    {
      theExpressions = c;
      return this;
    }

  TryExpr *AddClause(OnExceptionClause *c)
    {
      theClauses.append(c);
      return this;
    }
  TryExpr *PrependClause(OnExceptionClause *c)
    {
      theClauses.prepend(c);
      return this;
    }

  virtual EValue evalExpr()
    {EValue t; return t;};

  //RWS constructor
  virtual int output(int tc);
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer);
  virtual int walk(Symtab *context);
};


class LocalDecl : public Expression
{
  InitClauseList theInitClauses;
public:

  LocalDecl(InitClause *c)
    {
      theInitClauses.append(c);
      lineNo = lineno;
    }

  LocalDecl *AddInitClause(InitClause *c)
    {
      theInitClauses.append(c);
      return this;
    }
  LocalDecl *PrependInitClause(InitClause *c)
    {
      theInitClauses.prepend(c);
      return this;
    }

  virtual EValue evalExpr()
    {EValue t; return t;};

  //RWS need to fill in a constructor for list
  virtual int output(int tc);
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer);
  virtual int walk(Symtab *context);
};


class ConstantInitClause : public Production
{
  Symbol * theSymbol;
  Expression *theExpression;
  int isMagic;
public:
  ConstantInitClause(Symbol *sval,Expression *eval)
    {
      theSymbol=sval;
      theExpression=eval;
      lineNo=lineno;
      isMagic = 0;
    }
  void MarkConstant(void)
    {theSymbol->SetConstant();}
  virtual int output(int tc);
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer);
  virtual int walk(Symtab *context);
  int MarkMagic() {isMagic = 1;}
};

typedef DLList<ConstantInitClause *> CIClauseList;

class ConstantDecl : public Expression
{
  CIClauseList theDeclClauses;
  int isMagic;
public:

  ConstantDecl(ConstantInitClause *c)
    {
      theDeclClauses.append(c);
      lineNo = lineno;
      isMagic = 0;
    }

  MarkMagic() {isMagic = 1;}

  ConstantDecl *AddConstantClause(ConstantInitClause *c)
    {
      theDeclClauses.append(c);
      return this;
    }
  ConstantDecl *PrependConstantClause(ConstantInitClause *c)
    {
      theDeclClauses.prepend(c);
      return this;
    }

  virtual int output(int tc);
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer);
  virtual int walk(Symtab *context);
  virtual EValue evalExpr()
    {EValue t; return t;};
};


class ReturnExpr : public Expression
{
  Expression *theExpression;
public:
  ReturnExpr(Expression *val):theExpression(val),lineNo(lineno){}
  virtual int output(int tc);
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer);
  virtual int walk(Symtab *context);
  virtual EValue evalExpr()
    {EValue t; return t;};
};


class NString : public SimpleLiteral
{
  char *value;
public:
  NString(char *nv):value(nv),lineNo(lineno) {}
  virtual int output(int tc);
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer);
  virtual int walk(Symtab *context);
  virtual EValue evalExpr()
    {EValue t(value); return t;};
};

class NInteger : public SimpleLiteral
{
  int value;
public:
  NInteger(int thEValue):value(thEValue),lineNo(lineno){}
  virtual int output(int tc);
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer);
  virtual int walk(Symtab *context);
  virtual EValue evalExpr()
    {EValue t(value); return t;};
};

class NReal : public SimpleLiteral
{
  double value;
public:
  NReal(double thEValue):value(thEValue),lineNo(lineno){}

  virtual int output(int tc);
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer);
  virtual int walk(Symtab *context);
  virtual EValue evalExpr()
    {EValue t(value); return t;};
};

class NCharacter : public SimpleLiteral
{
  int value;
public:
  NCharacter(int thEValue):value(thEValue),lineNo(lineno) {}
  virtual int output(int tc);
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer);
  virtual int walk(Symtab *context);
  virtual EValue evalExpr()
    {EValue t(value); return t;};
};

class NTrue : public SimpleLiteral
{
public:
  NTrue():lineNo(lineno){}  
  virtual int output(int tc);
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer);
  virtual int walk(Symtab *context);
  virtual EValue evalExpr()
    {EValue t; return t;};
};

class NNil : public SimpleLiteral
{
public:
  NNil():lineNo(lineno){}  
  virtual int output(int tc);
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer);
  virtual int walk(Symtab *context);
  virtual EValue evalExpr()
    {EValue t; return t;};
};

class MPointer : public SimpleLiteral
{
  unsigned long value;
public:
  MPointer(unsigned long nv):value(nv),lineNo(lineno){}
  virtual int output(int tc);
  virtual int emit(Codeblock *theCode,NO_Frame *theContainer);
  virtual int walk(Symtab *context);
  virtual EValue evalExpr()
    {EValue t((unsigned long)value); return t;};
};





#endif
